unit FlyOverMenuSpdBtn;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Buttons, Menus, ExtCtrls;

type
  TMenuPosition = (posBottomRight,
                   posBottomLeft,
                   posTopRight,
                   posTopLeft,
                   posMiddleRight,
                   PosMiddleLeft);

  TFlyOverMenu = class(TCustomControl)
  private
    FMaxMenuItemHeight: integer;
    FMaxMenuItemWidth: integer;
    FLastMenuItemIndex: integer;
    FMenuItemIndex: integer;
    FMouseInControl: boolean;
    FOwner: TSpeedButton;
    FMenuPosition: TMenuPosition;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
//    procedure DrawMenuHeader(Value: string);
    function MenuIndexAtPoint(Pnt: TPoint): integer;
    function MenuItemRect(Index: integer): TRect;
    function MenuExists: boolean;
    function TotalMenuItemCount: integer;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    function MenuRect: TRect;
    procedure ActivateAtPoint(X, Y: integer);
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
  end;

  TFlyOverMenuSpdBtn = class(TSpeedButton)
  private
    FTimer: TTimer;
    FFlyOverMenu: TFlyOverMenu;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure TimerEvent(Sender: TObject);
    procedure StartTimer(Interval: integer);
    procedure StopTimer;
  protected
    procedure DoContextPopup(MousePos: TPoint; var Handled: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click; override;
  end;

procedure Register;

implementation

{$R *.RES}

procedure Register;
begin
  RegisterComponents('Samples', [TFlyOverMenuSpdBtn]);
end;

{ TFlyOverMenu }

//------------------------------------------------------------------------------
constructor TFlyOverMenu.Create(AOwner: TComponent);
begin
  inherited;
  FOwner             := TSpeedButton(AOwner);
  Color              := clMenu;
  Canvas.Font        := Screen.MenuFont;
  FLastMenuItemIndex := -1;
  FMenuItemIndex     := -1;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenu.ActivateAtPoint(X, Y: integer);
var
  Rct: TRect;
  RctWidth, RctHeight: integer;
begin
  if Assigned(FOwner.PopupMenu.OnPopup) then
    FOwner.PopupMenu.OnPopup(FOwner.PopupMenu);

  FLastMenuItemIndex := -1;
  FMenuItemIndex     := -1;

  Rct := MenuRect;
  OffsetRect(Rct, X, Y);

  RctWidth  := Rct.Right  - Rct.Left;
  RctHeight := Rct.Bottom - Rct.Top;

  if Rct.Bottom <= Screen.DesktopHeight then begin
    if Rct.Right < Screen.DesktopWidth then
      FMenuPosition := posBottomRight                         // 1.
    else begin
      FMenuPosition := posBottomLeft;                         // 2.
      OffsetRect(Rct, FOwner.Width - RctWidth, 0);
    end;
  end else begin
    OffsetRect(Rct, 0, - (FOwner.Height + RctHeight));
    if Rct.Top >= 0 then begin
      if Rct.Right < Screen.DesktopWidth then
        FMenuPosition := posTopRight                          // 3.
      else begin
        FMenuPosition := posTopLeft;                          // 4.
        OffsetRect(Rct, FOwner.Width - RctWidth, 0);
      end;
    end else begin
      OffsetRect(Rct, FOwner.Width, -Rct.Top);
      if Rct.Bottom > Screen.DesktopHeight then begin
        Rct.Bottom := Screen.DesktopHeight;
        RctHeight := Rct.Bottom - Rct.Top;
      end;
      if Rct.Right < Screen.DesktopWidth then
        FMenuPosition := posMiddleRight                       // 5.
      else begin
        FMenuPosition := posMiddleLeft;                       // 6.
        OffsetRect(Rct, - (FOwner.Width + RctWidth + 1), 0);
      end;
    end
  end;

  SetWindowPos(Handle,
               HWND_TOPMOST,
               Rct.Left,
               Rct.Top,
               RctWidth,
               RctHeight,
               SWP_SHOWWINDOW or SWP_NOACTIVATE);
  Invalidate;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenu.CMMouseEnter(var Message: TMessage);
begin
  FMouseInControl := True;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenu.CMMouseLeave(var Message: TMessage);
begin
  FMouseInControl := False;
  DestroyHandle;
end;

(*
//------------------------------------------------------------------------------
procedure TFlyOverMenu.DrawMenuHeader(Value: string);
var
  Rct: TRect;
begin
  Rct := Bounds(3, 3, FMaxMenuItemWidth, FMaxMenuItemHeight);
  Canvas.Brush.Color       := clMenu;
  Canvas.Font       := Screen.MenuFont;
  Canvas.Font.Color := clNavy;
  DrawText(Canvas.Handle, PChar(Value), Length(Value), Rct, DT_CENTER or DT_SINGLELINE);
  Rct := Bounds(3, FMaxMenuItemHeight, FMaxMenuItemWidth, 2);
  DrawEdge(Canvas.Handle, Rct, BDR_SUNKENOUTER, BF_RECT);
end;
*)

//------------------------------------------------------------------------------
function TFlyOverMenu.MenuExists: boolean;
begin
  Result := Assigned(FOwner.PopupMenu) and (FOwner.PopupMenu.Items.Count > 0);
end;

//------------------------------------------------------------------------------
function TFlyOverMenu.TotalMenuItemCount: integer;
begin
  Result := 0;
  if not MenuExists then
    Exit;

  Result := FOwner.PopupMenu.Items.Count;

//  if FOwner.ShowHint and (FOwner.Hint <> '') then
//    Inc(Result);
end;

//------------------------------------------------------------------------------
function TFlyOverMenu.MenuRect: TRect;
var
  I: integer;
  MaxWidth: integer;
  MaxHeight: integer;
begin
  FMaxMenuItemWidth  := 0;
  FMaxMenuItemHeight := 0;

  Result := Rect(0, 0, 0, 0);
  if not MenuExists then
    Exit;
  for I := 0 to FOwner.PopupMenu.Items.Count - 1 do begin
    MaxWidth  := Canvas.TextWidth(FOwner.PopupMenu.Items[I].Caption);
    MaxHeight := Canvas.TextHeight(FOwner.PopupMenu.Items[I].Caption);

    if MaxWidth > FMaxMenuItemWidth then
      FMaxMenuItemWidth := MaxWidth;

    if MaxHeight > FMaxMenuItemHeight then
      FMaxMenuItemHeight := MaxHeight;
  end;

(*
  if FOwner.ShowHint then begin
    MaxWidth := Canvas.TextWidth(FOwner.Hint);
    if MaxWidth > FMaxMenuItemWidth then
      FMaxMenuItemWidth := MaxWidth;
  end;
*)

  Inc(FMaxMenuItemWidth,  16);
  Inc(FMaxMenuItemHeight, 4);
  Result := Bounds(0, 0, FMaxMenuItemWidth + 6, FMaxMenuItemHeight * TotalMenuItemCount + 6);
end;

//------------------------------------------------------------------------------
function TFlyOverMenu.MenuIndexAtPoint(Pnt: TPoint): integer;
var
  Rct: TRect;
  I: integer;
begin
  Result := -1;
  Rct := Bounds(3, 3, FMaxMenuItemWidth, FMaxMenuItemHeight);
  if TotalMenuItemCount > FOwner.PopupMenu.Items.Count then
    OffsetRect(Rct, 0, FMaxMenuItemHeight);

  for I:=0 to FOwner.PopupMenu.Items.Count - 1 do begin
    if PtInRect(Rct, Pnt) then begin
      Result := I;
      Exit;
    end;
    OffsetRect(Rct, 0, FMaxMenuItemHeight);
  end;
end;

//------------------------------------------------------------------------------
function TFlyOverMenu.MenuItemRect(Index: integer): TRect;
begin
  Result := Bounds(3, 3, FMaxMenuItemWidth, FMaxMenuItemHeight);
  if TotalMenuItemCount > FOwner.PopupMenu.Items.Count then
    OffsetRect(Result, 0, FMaxMenuItemHeight);

  OffsetRect(Result, 0, FMaxMenuItemHeight * Index);
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenu.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  FMenuItemIndex := MenuIndexAtPoint(Point(X, Y));
  if FMenuItemIndex >= 0 then begin
    DestroyHandle;
    if Assigned(FOwner.PopupMenu.Items[FMenuItemIndex].OnClick) then
      FOwner.PopupMenu.Items[FMenuItemIndex].OnClick(FOwner.PopupMenu.Items[FMenuItemIndex]);
  end;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenu.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  FMenuItemIndex := MenuIndexAtPoint(Point(X, Y));
  if FMenuItemIndex <> FLastMenuItemIndex then begin
    FLastMenuItemIndex := FMenuItemIndex;
    // call InvalidateRect instead of Invalidate to prevent flicker
    InvalidateRect(Handle, nil, False);
  end;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenu.Paint;
const
  MenuItemOwnerDrawState: array[boolean] of TOwnerDrawState = ([odFocused],
                                                               [odSelected]);
var
  //  Hdr: string;
  Rct: TRect;
  I: integer;
begin
  inherited;
  if not MenuExists then
    Exit;

    Rct := ClientRect;
    DrawEdge(Canvas.Handle, Rct, BDR_RAISEDOUTER + BDR_RAISEDINNER, BF_RECT);

  (*
    Hdr := '';
    if FOwner.ShowHint then
      Hdr := FOwner.Hint;

    if Hdr <> '' then
      DrawMenuHeader(Hdr);
  *)

  Canvas.Font := Screen.MenuFont;
  for I:=0 to FOwner.PopupMenu.Items.Count - 1 do
    DrawMenuItem(FOwner.PopupMenu.Items[I],
                 Canvas,
                 MenuItemRect(I),
                 MenuItemOwnerDrawState[I = FMenuItemIndex]);
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenu.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := WS_POPUP;
    WindowClass.Style := WindowClass.Style or CS_SAVEBITS;
    if NewStyleControls then
      ExStyle := WS_EX_TOOLWINDOW;
  end;
end;

{ TFlyOverMenuSpdBtn }

//------------------------------------------------------------------------------
constructor TFlyOverMenuSpdBtn.Create(AOwner: TComponent);
begin
  inherited;
  FFlyOverMenu := TFlyOverMenu.Create(Self);
end;

//------------------------------------------------------------------------------
destructor TFlyOverMenuSpdBtn.Destroy;
begin
  StopTimer;
  inherited;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenuSpdBtn.Click;
begin
  inherited;
  FFlyOverMenu.DestroyHandle;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenuSpdBtn.CMMouseEnter(var Message: TMessage);
var
  Pnt: TPoint;
begin
  inherited;
  if Enabled then begin
    Pnt := ClientToScreen(Point(0, Height));
    FFlyOverMenu.ActivateAtPoint(Pnt.X, Pnt.Y);
  end;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenuSpdBtn.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if Enabled then
    StartTimer(10);
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenuSpdBtn.DoContextPopup(MousePos: TPoint; var Handled: Boolean);
begin
  inherited;
  Handled := True;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenuSpdBtn.StartTimer(Interval: integer);
begin
  StopTimer;
  FTimer := TTimer.Create(Self);
  FTImer.Interval := Interval;
  FTimer.OnTimer := TimerEvent;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenuSpdBtn.StopTimer;
begin
  if Assigned(FTimer) then begin
    FTimer.Enabled := False;
    FreeAndNil(FTimer);
  end;
end;

//------------------------------------------------------------------------------
procedure TFlyOverMenuSpdBtn.TimerEvent(Sender: TObject);
begin
  StopTimer;
  if not FFlyOverMenu.FMouseInControl then
    FFlyOverMenu.DestroyHandle;
end;

end.
